"""
   Copyright 2022 Nik Dennler

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from utils.analysis import *

# Define data directory and get data loader
data_dir_parent = Path("./src/data/WTD_upload/")

# Define and create output directory
output_dir = Path("./src/output/test_runs")
output_dir_experiment = output_dir.joinpath("01_experimental_protocol")
output_dir_baseline_drift = output_dir.joinpath("02_baseline_drift")
output_dir_full = output_dir.joinpath("03_full_analysis")
output_dir_restricted = output_dir.joinpath("04_restricted_analysis")

# Select Hyperparams (same for all analysis)
loc_dir = 4  # 1.18m from source
volt = 5  # 6V sensor heating
speed = 2  # 1500rpm fan speed

# Select Data Time Windows
ts_start_pca = np.array([0.01, 10.0, 20.0, 30.0, 40.0, 50.0]) * 1000
ts_start_classifier = (np.array([0.01] + list(range(5, 101, 5)))) * 1000  # in ms
dt_window = 0.1 * 1000  # in ms

# Define experimental parameters
consider_resistance = True  # False
train_ratio = 0.8
n_experiments = 10  # For Classifier
load_previous = True  # Set this to False if you want to run spacial distribution analysis and SVM classifier again

run_experiments = {
    "indicate_windtunnel_location_full": True,
    "indicate_windtunnel_location_restricted": True,
    "single_run": True,
    "gas_id_analysis": True,
    "baseline_plot": True,
    "spacial_distribution_analysis_shortterm": True,
    "spacial_distribution_analysis_longterm": True,
    "pca_analysis_full_raw": True,
    "pca_analysis_full_offsetsubtracted": True,
    "pca_analysis_restriced_raw": False,
    "pca_analysis_restricted_offsetsubtracted": False,
    "svm_classifier_full": True,
    "svm_classifier_restricted": True,
}

# Get data and create output directories
data = DataSelector(
    str(data_dir_parent),
    drop_duplicates=True,
    resample=True,
    verbose=False,
    use_HDFcache=True,
)

output_dir_experiment.mkdir(exist_ok=True, parents=True)
output_dir_baseline_drift.mkdir(exist_ok=True, parents=True)
output_dir_full.mkdir(exist_ok=True, parents=True)
output_dir_restricted.mkdir(exist_ok=True, parents=True)

# Run experiments
if run_experiments["indicate_windtunnel_location_full"]:
    print("INDICATE WINDTUNNEL LOCATION FULL")
    visualize.indicate_windtunnel_location(
        output_dir=output_dir_full, location=loc_dir, modules=constants.modules_full
    )

if run_experiments["indicate_windtunnel_location_restricted"]:
    print("INDICATE WINDTUNNEL LOCATION RESTRICED")
    visualize.indicate_windtunnel_location(
        output_dir=output_dir_restricted,
        location=loc_dir,
        modules=constants.modules_restricted,
    )

if run_experiments["single_run"]:
    print("DISPLAY SINGLE RUN")
    single_experiment(
        data,
        volt,
        loc=loc_dir,
        speed=speed,
        gas_int=1,
        module=constants.modules_full[0],
        columns=constants.columns_all,
        crop_in_time=False,
        consider_resistance=consider_resistance,
        output_dir=output_dir_experiment,
        show=False,
    )

if run_experiments["gas_id_analysis"]:
    print("GAS ID ANALYSIS")
    gas_id_analysis(data_dir_parent, output_dir=output_dir_experiment)

if run_experiments["baseline_plot"]:
    print("DISPLAY BASELINE VALUES")
    baseline_plot_onerow(
        data,
        loc_dir,
        volt,
        speed,
        constants.columns_full,
        subtracting_dc_offset=False,
        consider_resistance=consider_resistance,
        output_dir=output_dir_baseline_drift,
    )

if run_experiments["spacial_distribution_analysis_shortterm"]:
    print("SPACIAL ANALYSIS OF SHORT-TERM VARIATIONS")
    spacial_distribution_analysis_shortterm(
        data,
        volt,
        speed,
        output_dir=output_dir_baseline_drift,
        load_previous=load_previous,
        consider_resistance=consider_resistance,
    )

if run_experiments["spacial_distribution_analysis_longterm"]:
    print("SPACIAL ANALYSIS OF LONG-TERM VARIATIONS")
    spacial_distribution_analysis_longterm(
        data,
        volt,
        speed,
        output_dir=output_dir_baseline_drift,
        load_previous=load_previous,
        consider_resistance=consider_resistance,
    )

if run_experiments["pca_analysis_full_raw"]:
    print("PCA ANALYSIS OF FULL DATASET, RAW")
    pca_analysis(
        data,
        ts_start_pca,
        dt_window,
        volt,
        speed,
        loc_dir,
        constants.columns_full,
        constants.skip_gas_full,
        subtracting_dc_offset=False,
        consider_resistance=consider_resistance,
        output_dir=output_dir_full,
    )

if run_experiments["pca_analysis_full_offsetsubtracted"]:
    print("PCA ANALYSIS OF FULL DATASET, OFFSET SUBTRACTED")
    pca_analysis(
        data,
        ts_start_pca,
        dt_window,
        volt,
        speed,
        loc_dir,
        constants.columns_full,
        constants.skip_gas_full,
        subtracting_dc_offset=True,
        consider_resistance=consider_resistance,
        output_dir=output_dir_full,
    )

if run_experiments["pca_analysis_restricted_raw"]:
    print("PCA ANALYSIS OF RESTRICTED DATASET, RAW")
    pca_analysis(
        data,
        ts_start_pca,
        dt_window,
        volt,
        speed,
        loc_dir,
        constants.columns_restricted,
        constants.skip_gas_restricted,
        subtracting_dc_offset=False,
        consider_resistance=consider_resistance,
        output_dir=output_dir_full,
    )

if run_experiments["pca_analysis_restricted_offsetsubtracted"]:
    print("PCA ANALYSIS OF RESTRICTED DATASET, OFFSET SUBTRACTED")
    pca_analysis(
        data,
        ts_start_pca,
        dt_window,
        volt,
        speed,
        loc_dir,
        constants.columns_restricted,
        constants.skip_gas_restricted,
        subtracting_dc_offset=True,
        consider_resistance=consider_resistance,
        output_dir=output_dir_full,
    )


if run_experiments["svm_classifier_full"]:
    print("SVM ANALYSIS OF FULL DATASET")
    simple_classifier(
        data,
        classifier="svm",
        loc_dir=loc_dir,
        volt=volt,
        speed=speed,
        columns=constants.columns_full,
        skip_gas=constants.skip_gas_full,
        ts_start=ts_start_classifier,
        dt_window=dt_window,
        train_ratio=train_ratio,
        n_experiments=n_experiments,
        load_previous=load_previous,  # False,
        consider_resistance=consider_resistance,
        output_dir=output_dir_full,
    )

if run_experiments["svm_classifier_restricted"]:
    print("SVM ANALYSIS OF RESTRICTED DATASET")
    simple_classifier(
        data,
        classifier="svm",
        loc_dir=loc_dir,
        volt=volt,
        speed=speed,
        columns=constants.columns_restricted,
        skip_gas=constants.skip_gas_restricted,
        ts_start=ts_start_classifier,
        dt_window=dt_window,
        train_ratio=train_ratio,
        n_experiments=n_experiments,
        load_previous=load_previous,  # False,
        consider_resistance=consider_resistance,
        output_dir=output_dir_restricted,
    )
