"""
   Copyright 2022 Nik Dennler

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
from pathlib import Path

# import tqdm
from tqdm.autonotebook import tqdm

import numpy as np
import pandas as pd
import matplotlib as mpl
from sklearn.decomposition import PCA
from sklearn import svm
from sklearn.naive_bayes import GaussianNB

from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

# from sklearn import svm

np_load_old = np.load
# modify the default parameters of np.load
np.load = lambda *a, **k: np_load_old(*a, allow_pickle=True, **k)

date64 = np.datetime64("2000-01-01")
import matplotlib as mpl

new_rc_params = {"text.usetex": False, "svg.fonttype": "none"}
mpl.rcParams.update(new_rc_params)

# from lib.bioml_toolbox.src.data_loaders.vergara_opensampling.utils import *
from utils.data_loader.utils import *
import utils.visualize as visualize
import utils.constants as constants


def _get_resistance(df, V_c=3110, R_l=10):
    """Calculates resistance from voltage

    Args:
        df ([type]): [description]
        V_c (int, optional): [description]. Defaults to 3110.
        R_l (int, optional): [description]. Defaults to 10.

    Returns:
        [type]: [description]
    """
    return (V_c - df) / df * R_l


def single_experiment(
    data,
    volt,
    gas_int,
    module,
    columns,
    loc=4,
    speed=2,
    crop_in_time=False,
    consider_resistance=True,
    output_dir=None,
    show=False,
):
    """Displaying single experimental trial

    Args:
        data (data_loader object): data loader
        volt (int): hotplate heating setting. value in [1,...,6]
        gas_int (int): gas setting. See utils.constants.GasNames for details
        columns (list): list of columns. in the form 'B2S4'
        loc (int): Longitudinal location in wind tunnel. Defaults to 4. 
        speed (int): Wind speed setting. Defaults to 2.
        crop_in_time (bool, optional): Defaults to False.
        consider_resistance (bool, optional): Defaults to True. If false, take sensor voltages
        output_dir ([type], optional): Defaults to None.
        show (bool, optional): Defaults to False.
    """
    sensor_column = data.select(
        gas=int(gas_int), loc=loc, voltage=volt, speed=speed, trial=int(1)
    )[0]
    df = sensor_column.get_data()
    df = df[columns]

    if crop_in_time:
        t_start = 0 * 1000
        t_end = 60 * 1000
        df = df[(df.index >= t_start) & (df.index <= t_end)]

    if consider_resistance:
        df = _get_resistance(df)

    visualize.single_experiment(
        df,
        gas_int=gas_int,
        module=module,
        consider_resistance=consider_resistance,
        output_dir=output_dir,
        show=show,
    )


def gas_id_analysis(data_dir_parent, output_dir=None, show=False):
    """Plot gas id versus timestamp

    Args:
        data_dir_parent (str): data directory
        output_dir (str, optional): output directory. Defaults to None.
        show (bool, optional): Defaults to False.
    """
    all_datetimes = []
    all_idx = []
    all_locs = []

    for idx, gas in zip(
        tqdm(constants.GasNames.keys(), desc="gas"), constants.GasNames.values()
    ):
        data_dir = data_dir_parent.joinpath(gas)
        for loc_idx, loc_dir in enumerate(sorted(data_dir.iterdir())):
            loc = loc_dir.name

            for file in sorted(loc_dir.iterdir()):
                datetimestr = file.name.split("_")[0]

                try:
                    datetime = pd.to_datetime(datetimestr)

                    all_datetimes.append(datetime)
                    all_idx.append(idx)
                    all_locs.append(int(loc[1]))

                except TypeError:
                    pass
    print("Visualize...")
    visualize.timestamp_vs_gasid(
        all_datetimes, all_idx, all_locs, output_dir=output_dir, show=show
    )


def baseline_plot_onerow(
    data,
    loc_dir,
    volt,
    speed,
    columns,
    subtracting_dc_offset=False,
    consider_resistance=True,
    output_dir=None,
    show=False,
):
    """For each trial, plotting averaged baseline over timestamp

    Args:
        data (data_loader object): data loader
        loc_dir (int): location setting. valie in [1,...,6]
        volt (int): hotplate heating setting. value in [1,...,6]
        speed (int): wind speed setting. value in [1,2,3]
        columns (list): list of columns. in the form 'B2S4'
        subtracting_dc_offset (bool, optional): Defaults to False.
        consider_resistance (bool, optional): [description]. Defaults to True.
        output_dir (str, optional): Defaults to None.
        show (bool, optional): [description]. Defaults to False.
    """
    t_start = 0.0 * 1000
    dt = 20 * 1000
    t_end = t_start + dt + constants.eps
    module = columns[0][1]
    sensors = [int(column[3]) for column in columns]

    all_datetimes = []
    all_baselines = []
    all_gases = []
    all_gas_ints = []

    gas_ints = np.array(list(constants.GasNames.keys()))

    for gas_int in tqdm(gas_ints, desc="gas"):
        gas = constants.GasNames[gas_int]
        # print(gas)
        for trial in constants.all_trials:
            try:
                sensor_column = data.select(
                    gas=int(gas_int),
                    loc=loc_dir,
                    voltage=volt,
                    speed=speed,
                    trial=int(trial),
                )[0]
            except IndexError:
                continue
            df = sensor_column.get_data()
            df = df[columns]

            if consider_resistance:
                df = _get_resistance(df)

            if subtracting_dc_offset:
                df = df - df.iloc[0]  # Subtracting DC offset

            df = df[(df.index >= t_start) & (df.index <= t_end)]  # Cropping in time

            filename = sensor_column._get_filename()
            datetime = pd.to_datetime(filename[:12])
            df_baseline = df.mean()

            all_datetimes.append(datetime)
            all_baselines.append(df_baseline)
            all_gases.append(gas)
            all_gas_ints.append(gas_int)
    all_datetimes, all_baselines, all_gases, all_gas_ints = zip(
        *sorted(zip(all_datetimes, all_baselines, all_gases, all_gas_ints))
    )

    print("Visualize...")
    visualize.time_vs_baseline(
        all_datetimes,
        all_baselines,
        all_gases,
        all_gas_ints,
        module=module,
        sensors=sensors,
        output_dir=output_dir,
        show=show,
    )


def spacial_distribution_analysis_longterm(
    data,
    volt,
    speed,
    output_dir=None,
    load_previous=True,
    consider_resistance=True,
    mode="Short",
    show=False,
):
    """Analyse spacial distribution, on sensor board and in wind tunnel.

    Args:
        data (data_loader object): data loader
        volt (int): hotplate heating setting. value in [1,...,6]
        speed (int): wind speed setting. value in [1,2,3]
        output_dir (str, optional): Defaults to None.
        load_previous (bool, optional): Run analysis or load previous. Defaults to True.
        consider_resistance (bool, optional): [description]. Defaults to True.
        mode (str, optional): Drift mode Either 'Short' or 'Long'. Defaults to 'Short'.
        show (bool, optional): Defaults to False.        
    """
    mode = "Long"

    if load_previous:
        try:
            all_baselines = np.load(output_dir.joinpath("all_baselines.npy"))
        except FileNotFoundError:
            print("Cannot load previous.")
            load_previous = False

    if not load_previous:

        t_start = 0 * 1000
        dt = 20 * 1000
        t_end = t_start + dt + constants.eps

        gas_ints = np.array(list(constants.GasNames.keys()))
        all_baselines = []
        for idx_loc, loc in enumerate(
            tqdm(constants.all_locs, desc=" locs", position=0)
        ):
            baselines_loc = []  # np.zeros(len(constants.all_boards))
            for idx_m, m in enumerate(
                tqdm(constants.all_boards, desc=" m", position=1, leave=False)
            ):
                print(f"L{loc}M{m}")
                baselines_loc_module = []
                columns = [f"B{m}S{s}" for s in constants.sensors_full]
                for gas_int in gas_ints:
                    gas = constants.GasNames[gas_int]
                    for trial in constants.all_trials:
                        try:
                            sensor_column = data.select(
                                gas=int(gas_int),
                                loc=loc,
                                voltage=volt,
                                speed=speed,
                                trial=int(trial),
                            )[0]
                        except (IndexError, AssertionError) as e:
                            continue
                        df = sensor_column.get_data()
                        df = df[columns]
                        df = df[
                            (df.index >= t_start) & (df.index <= t_end)
                        ]  # Crop in time

                        if consider_resistance:
                            df = _get_resistance(df)

                        df_baseline = df.mean().to_numpy()
                        baselines_loc_module.append(df_baseline)
                baselines_loc.append(np.array(baselines_loc_module))
            all_baselines.append(baselines_loc)
        all_baselines = np.array(all_baselines)
        np.save(output_dir.joinpath("all_baselines.npy"), all_baselines)
        all_baselines = np.load(output_dir.joinpath("all_baselines.npy"))

    std_long = np.array(
        [
            [np.std(all_baselines[i][j], axis=0) for j in range(len(all_baselines[0]))]
            for i in range(len(all_baselines))
        ]
    )
    mean_long = np.array(
        [
            [np.mean(all_baselines[i][j], axis=0) for j in range(len(all_baselines[0]))]
            for i in range(len(all_baselines))
        ]
    )
    coefficient_of_variance_long = std_long / mean_long

    l, m, s = coefficient_of_variance_long.shape
    sensor_drift = [0] + list(np.mean(coefficient_of_variance_long, axis=(0, 1)))
    all_drift = np.sqrt(np.mean(coefficient_of_variance_long, axis=2).T)
    all_sensors = np.reshape(coefficient_of_variance_long.T, (s, m * l)).T

    print("Visualize...")
    visualize.draw_windtunnel(all_drift, output_dir, name=mode, show=show)
    visualize.draw_boxplot_drift(all_sensors, output_dir, name=mode, show=show)
    visualize.draw_sensorboard(sensor_drift, output_dir, name=mode, show=show)


def spacial_distribution_analysis_shortterm(
    data,
    volt,
    speed,
    output_dir=None,
    load_previous=True,
    consider_resistance=True,
    mode="Short",
    show=False,
):
    """Analyse spacial distribution, on sensor board and in wind tunnel.

    Args:
        data_dir_parent (str): data directory
        data (data_loader object): data loader
        volt (int): hotplate heating setting. value in [1,...,6]
        speed (int): wind speed setting. value in [1,2,3]
        output_dir (str, optional): Defaults to None.
        load_previous (bool, optional): Run analysis or load previous. Defaults to True.
        consider_resistance (bool, optional): [description]. Defaults to True.
        mode (str, optional): Drift mode Either 'Short' or 'Long'. Defaults to 'Short'.
        show (bool, optional): Defaults to False.        
    """
    mode = "Short"

    if load_previous:
        print(output_dir)
        try:
            all_coefficients_of_variances = np.load(
                output_dir.joinpath("all_coefficients_of_variances.npy")
            )
        except FileNotFoundError:
            print("Cannot load previous.")
            load_previous = False

    if not load_previous:
        t_start = 0 * 1000
        dt = 20 * 1000
        t_end = t_start + dt + constants.eps

        ## SHORT TERM DRIFT: Variance of all baseline values
        print("short")
        gas_ints = np.array(list(constants.GasNames.keys()))
        all_variances = []
        all_means = []
        all_index_of_dispersion = []
        all_coefficients_of_variances = []
        for loc in tqdm(constants.all_locs, desc=" locs", position=0):
            all_variances_loc = []
            all_means_loc = []
            all_index_of_dispersion_loc = []
            all_coefficients_of_variances_loc = []
            for m in tqdm(constants.all_boards, desc=" m", position=1, leave=False):
                # print(f"L{loc}M{m}")
                columns = [f"B{m}S{s}" for s in constants.sensors_full]
                all_variances_loc_module = []
                all_means_loc_module = []
                all_index_of_dispersion_loc_module = []
                all_coefficients_of_variances_loc_module = []
                for gas_int in gas_ints:
                    gas = constants.GasNames[gas_int]
                    # print(gas)
                    for trial in constants.all_trials:
                        try:
                            sensor_column = data.select(
                                gas=int(gas_int),
                                loc=loc,
                                voltage=volt,
                                speed=speed,
                                trial=int(trial),
                            )[0]
                        except (IndexError, AssertionError) as e:
                            continue
                        df = sensor_column.get_data()
                        df = df[columns]

                        df = df[
                            (df.index >= t_start) & (df.index <= t_end)
                        ]  # Crop in time
                        if consider_resistance:
                            df = _get_resistance(df)

                        variance = (((df - df.mean()) ** 2).sum()) / df.count()
                        coefficient_of_variance = np.sqrt(variance) / df.mean()
                        index_of_dispersion = variance / df.mean()

                        all_variances_loc_module.append(variance.to_numpy())
                        all_means_loc_module.append(df.mean().to_numpy())
                        all_coefficients_of_variances_loc_module.append(
                            coefficient_of_variance.to_numpy()
                        )
                        all_index_of_dispersion_loc_module.append(
                            index_of_dispersion.to_numpy()
                        )

                all_variances_loc.append(
                    np.array(all_variances_loc_module).mean(axis=0)
                )
                all_means_loc.append(np.array(all_means_loc_module).mean(axis=0))
                all_coefficients_of_variances_loc.append(
                    np.array(all_coefficients_of_variances_loc_module).mean(axis=0)
                )
                all_index_of_dispersion_loc.append(
                    np.array(all_index_of_dispersion_loc_module).mean(axis=0)
                )

            all_variances.append(all_variances_loc)
            # print(all_variances)
            all_means.append(all_means_loc)
            all_coefficients_of_variances.append(all_coefficients_of_variances_loc)
            all_index_of_dispersion.append(all_index_of_dispersion_loc)

        all_variances = np.array(all_variances)
        all_means = np.array(all_means)
        all_coefficients_of_variances = np.array(all_coefficients_of_variances)
        all_index_of_dispersion = np.array(all_index_of_dispersion)

        # np.save(output_dir.joinpath("all_variances.npy"), all_variances)
        # np.save(output_dir.joinpath("all_means.npy"), all_means)
        np.save(
            output_dir.joinpath("all_coefficients_of_variances.npy"),
            all_coefficients_of_variances,
        )
        # np.save(
        #     output_dir.joinpath("all_index_of_dispersion.npy"), all_index_of_dispersion
        # )
        print("Done short")

    l, m, s = all_coefficients_of_variances.shape
    sensor_drift = [0] + list(np.mean(all_coefficients_of_variances, axis=(0, 1)))
    all_drift = np.mean(all_coefficients_of_variances, axis=2).T
    all_sensors = np.reshape(all_coefficients_of_variances.T, (s, m * l)).T

    print("Visualize...")
    visualize.draw_windtunnel(all_drift, output_dir, name=mode, show=show)
    visualize.draw_boxplot_drift(all_sensors, output_dir, name=mode, show=show)
    visualize.draw_sensorboard(sensor_drift, output_dir, name=mode, show=show)


def pca_analysis(
    data,
    ts_start,
    dt_window,
    volt,
    speed,
    loc_dir,
    columns,
    skip_gas,
    subtracting_dc_offset=False,
    consider_resistance=True,
    output_dir=None,
    show=False,
):
    """Run Principal Component Analysis (PCA) for different time windows

    Args:
        data (data_loader object): data loader
        ts_start (list): List of starting times, in milliseconds.
        dt_window ([type]): Size of time window, in milliseconds
        volt (int): hotplate heating setting. value in [1,...,6]
        speed (int): wind speed setting. value in [1,2,3]
        loc_dir (int): location setting. valie in [1,...,6]
        columns (list): list of columns. in the form 'B2S4'
        subtracting_dc_offset (bool, optional): Defaults to False.
        consider_resistance (bool, optional): [description]. Defaults to True.
        output_dir (str, optional): Defaults to None.
        show (bool, optional): [description]. Defaults to False.    
    """
    X_train_all = []
    for i, t_start in enumerate(ts_start):
        print(f"t_start: {t_start/1000}s")
        t_end = t_start + dt_window + constants.eps

        X_train = []
        y_train = []
        gas_ints = np.array(list(constants.GasNames.keys()))
        for gas_int in tqdm(gas_ints, desc="gas"):
            gas = constants.GasNames[gas_int]
            if gas in skip_gas:
                continue
            # print(gas)
            for trial in constants.all_trials:
                sensor_column = data.select(
                    gas=int(gas_int),
                    loc=loc_dir,
                    voltage=volt,
                    speed=speed,
                    trial=int(trial),
                )[0]
                df = sensor_column.get_data()
                df = df[columns]

                if consider_resistance:
                    df = _get_resistance(df)

                if subtracting_dc_offset:
                    df = df - np.mean(df.loc[0:dt_window])

                # Crop in time
                df = df[(df.index >= t_start) & (df.index <= t_end)]

                X_temp = df.to_numpy().flatten()  # assuming IID features
                X_train.append(X_temp)
                y_train.append(gas_int)

        X_train_all.append(X_train)
    X_train_all = np.array(X_train_all)

    outer, inner, innerst = X_train_all.shape
    X_train_all_flattened = X_train_all.reshape((outer * inner, innerst))

    print("PCA")
    pca = PCA()
    pca.fit(X_train_all_flattened)  # Spanning space of Principal Components

    print("Visualize...")
    visualize.pca_projections(
        pca,
        ts_start,
        X_train_all,
        y_train,
        subtracting_dc_offset,
        consider_resistance,
        output_dir=output_dir,
        show=show,
    )


def simple_classifier(
    data,
    classifier,
    loc_dir,
    volt,
    speed,
    columns,
    skip_gas,
    ts_start,
    dt_window,
    train_ratio=0.8,
    n_experiments=10,
    load_previous=True,
    consider_resistance=True,
    output_dir=None,
    show=False,
):
    """Simple classifier for classifying time windows

    Args:
        data (data_loader object): data loader
        classifier (str): Classifier. Choose either 'svm' or 'naivebayes'
        loc_dir (int): location setting. valie in [1,...,6]
        volt (int): hotplate heating setting. value in [1,...,6]
        speed (int): wind speed setting. value in [1,2,3]
        columns (list): list of columns. in the form 'B2S4'
        ts_start (list): List of starting times, in milliseconds.
        dt_window ([type]): Size of time window, in milliseconds
        train_ratio (float): Ratio between train and validation set. Defaults to 0.8. 
        n_experiments (int): Number of experiments. Defaults to 10. 
        load_previous (bool, optional): Run analysis or load previous. Defaults to True.
        consider_resistance (bool, optional): Defaults to True.
        output_dir (str, optional): Defaults to None.
        show (bool, optional): [description]. Defaults to False.  
    """
    if load_previous:
        if classifier == "svm":
            try:
                all_accs_dc = np.load(
                    output_dir.joinpath(
                        f"all_accs_dc_resistance{consider_resistance}.npy"
                    )
                )
                all_accs_gain = np.load(
                    output_dir.joinpath(
                        f"all_accs_gain_resistance{consider_resistance}.npy"
                    )
                )
            except FileNotFoundError:
                print("Cannot load previous.")
                load_previous = False

    if not load_previous:
        # Train-Val-Test Split
        n_all = 20
        n_test = 0
        n_val = int(round((1 - train_ratio) * n_all))
        print(n_val)
        n_train = n_all - n_test - n_val

        for subtracting_dc_offset in [True, False]:
            print(f"zero-offset subtraction: {subtracting_dc_offset}")
            trials_all = range(1, n_all + 1)

            np.random.seed(42)
            trials_test = list(np.random.choice(trials_all, n_test, replace=False))

            all_accs_linsvm_ovo = []

            for exp in tqdm(range(1, n_experiments + 1)):
                randint = np.random.randint(
                    10000
                )  # Ensuring same train/test split for all windows of same experiment

                accs_linsvm_ovo = []

                for i, t_start in enumerate(ts_start):
                    np.random.seed(randint)

                    print(f"t_start: {t_start}")
                    t_end = t_start + dt_window + constants.eps

                    trials_val = list(
                        np.random.choice(
                            np.setdiff1d(trials_all, trials_test), n_val, replace=False
                        )
                    )
                    trials_train = list(
                        np.random.choice(
                            np.setdiff1d(trials_all, trials_val + trials_test),
                            n_train,
                            replace=False,
                        )
                    )
                    print("trials train: ", trials_train, ", trials val: ", trials_val)

                    X_train = []
                    y_train = []
                    X_val = []
                    y_val = []
                    gas_ints = np.array(list(constants.GasNames.keys()))
                    for gas_int in gas_ints:
                        gas = constants.GasNames[gas_int]
                        if gas in skip_gas:
                            continue

                        for trial in trials_train:
                            sensor_column = data.select(
                                gas=int(gas_int),
                                loc=loc_dir,
                                voltage=volt,
                                speed=speed,
                                trial=int(trial),
                            )[0]
                            df = sensor_column.get_data()
                            df = df[columns]

                            if consider_resistance:
                                df = _get_resistance(df)

                            if subtracting_dc_offset:
                                df = df - np.mean(df.loc[0:dt_window])

                            df = df[
                                (df.index >= t_start) & (df.index <= t_end)
                            ]  # Crop in time

                            X_temp = df.to_numpy().flatten()  # assuming IID features
                            X_train.append(X_temp)
                            y_train.append(gas_int)

                        for trial in trials_val:
                            sensor_column = data.select(
                                gas=int(gas_int),
                                loc=loc_dir,
                                voltage=volt,
                                speed=speed,
                                trial=int(trial),
                            )[0]
                            df = sensor_column.get_data()
                            df = df[columns]

                            if consider_resistance:
                                df = _get_resistance(df)

                            if subtracting_dc_offset:
                                df = df - np.mean(df.loc[0:dt_window])

                            # Crop in time
                            df = df[(df.index >= t_start) & (df.index <= t_end)]

                            X_temp = df.to_numpy().flatten()  # assuming IID features
                            X_val.append(X_temp)
                            y_val.append(gas_int)

                    X_train = np.array(X_train)
                    X_val = np.array(X_val)
                    y_train = np.array(y_train)
                    y_val = np.array(y_val)

                    # Linear SVM OVO
                    linear = svm.SVC(
                        kernel="linear", C=1, decision_function_shape="ovo"
                    ).fit(X_train, y_train)

                    accuracy_lin_ovo = linear.score(X_val, y_val)
                    accs_linsvm_ovo.append(accuracy_lin_ovo)
                    print("Linear SVM (One vs One), Accuracy:", accuracy_lin_ovo)

                all_accs_linsvm_ovo.append(accs_linsvm_ovo)

            all_accs_linsvm_ovo = np.array(all_accs_linsvm_ovo)

            np.save(
                output_dir.joinpath(
                    f"all_accs_linearsvm_ovo_dcoffsetsubtr{subtracting_dc_offset}_resistance{consider_resistance}.npy"
                ),
                all_accs_linsvm_ovo,
            )

            if not subtracting_dc_offset:
                if classifier == "svm":
                    all_accs_dc = all_accs_linsvm_ovo
            else:
                if classifier == "svm":
                    all_accs_gain = all_accs_linsvm_ovo

    print("Visualize...")
    visualize.classification_accuracies(
        ts_start,
        all_accs_dc,
        all_accs_gain,
        skip_gas,
        classifier="SVM",
        consider_resistance=consider_resistance,
        output_dir=output_dir,
        show=show,
    )

    return np.mean(all_accs_dc, axis=0), np.mean(all_accs_gain, axis=0)
