"""
   Copyright 2022 Nik Dennler

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.pyplot import cm
from matplotlib.patches import Rectangle, Circle
from matplotlib.colors import Normalize
from matplotlib import dates as mPlotDATEs
import matplotlib as mpl

new_rc_params = {"text.usetex": False, "svg.fonttype": "none"}
mpl.rcParams.update(new_rc_params)

import utils.constants as constants

cm_inferno = plt.cm.inferno
cm_inferno_r = plt.cm.inferno_r
cm_viridis = plt.cm.viridis
cm_coolwarm = plt.cm.coolwarm
cm_rainbow = plt.cm.rainbow
cm_reds = plt.cm.OrRd
cm_magma = plt.cm.inferno


def _get_nestedeventtimeslist_and_locs(times, values, others=None):
    """Get structure suitable for plt.eventplot(). Converting list of times and list of values to eventtimes and locations.

    Args:
        times (list): list with times of event
        values (list): list with value/category of event

    Returns:
        (tuple of two lists): nested list of eventtimes and list of all locs
    """
    all_times = []
    all_others = []

    unique_values = sorted(np.unique(values))
    for l in unique_values:
        all_times.append([])
        all_others.append([])

    for idx, unq in enumerate(unique_values):
        if others is None:
            for v, t in zip(values, times):
                if v == unq:
                    all_times[idx].append(mPlotDATEs.date2num(t.to_pydatetime()))
        else:
            for v, t, o in zip(values, times, others):
                if v == unq:
                    all_times[idx].append(mPlotDATEs.date2num(t.to_pydatetime()))
                    all_others[idx].append(o)

        all_times[idx], all_others[idx] = [
            list(t) for t in zip(*sorted(zip(all_times[idx], all_others[idx])))
        ]

    return all_times, unique_values, all_others


def _normalize(value, min, max):
    """Just basic normalization.

    Args:
        value (float object): [description]
        min (float): [description]
        max (float): [description]

    Returns:
        float object: (value-min)/(max-min)
    """
    return (value - min) / (max - min)


def _get_unique_labels(fig, axes=None):
    """Gets unique labels from mpl.pyplot.figure object (removes duplicate labels for legend). If axes is not None, considers those istead of whole figure.

    Args:
        fig (mpl.pyplot object): your figure
        axes (list of ax object, optional): Your list of ax objects. Defaults to None -> considers whole figure.

    Returns:
        list: list of lines object for legend.
    """
    lines = []
    labels = []
    if axes is None:
        for ax in fig.axes:
            axLine, axLabel = ax.get_legend_handles_labels()
            lines.extend(axLine)
            labels.extend(axLabel)
    else:
        for ax in axes:
            axLine, axLabel = ax.get_legend_handles_labels()
            lines.extend(axLine)
            labels.extend(axLabel)
    labels, lines = zip(*sorted(zip(labels, lines), key=lambda t: t[0]))
    unique = [
        (h, l) for i, (h, l) in enumerate(zip(lines, labels)) if l not in labels[:i]
    ]
    return unique


def single_experiment(
    df, gas_int=1, module=5, consider_resistance=True, output_dir=None, show=False
):
    """Visualize single experiment

    Args:
        df (Pandas DataFrame): DataFrame
        gas_int (int, optional): Integer hashing the gas. Defaults to 1.
        module (int, optional): Integer hashing the lateral position (the board module). Defaults to 5.
        consider_resistance (bool, optional): Plot resistance rather than direct sensor value. Defaults to True.
        output_dir (str, optional): Output directory. Defaults to None.
        show (bool, optional): Show plot. Defaults to False.
    """
    # Plot
    fig, ax = plt.subplots(figsize=(8, 5))
    ax.axvspan(20, 200, alpha=0.2, color=cm_rainbow(gas_int / 11))
    for i in constants.sensors_all:
        ax.plot(
            df.index / 1000,
            df[f"B{module}S{i}"],
            color=cm_viridis((i - 2) / 7),
            label=f"S{i}",
        )

    # Format
    if consider_resistance:
        ax.set_ylabel(r"$R_{sensor}$ ($k\Omega$)")
    else:
        ax.set_ylabel(r"$V_{sensor}$ ($mV$)")
    ax.set_xlabel(r"Time ($s$)")
    xticks = list(ax.get_xticks())
    xticks.append(20)
    ax.set_xticks(xticks)
    xmin = min(df.index / 1000 - 1)
    xmax = max(df.index / 1000)
    ax.set_xlim([xmin, xmax])
    ax.legend(loc="upper right", ncol=2, framealpha=0.85, fancybox=True)

    # Save&Show
    if output_dir is not None:
        plt.savefig(output_dir.joinpath("1b_single_gas.png"), dpi=600, bbox_inches="tight")
        plt.savefig(output_dir.joinpath("1b_single_gas.svg"), bbox_inches="tight")
    if show:
        plt.show()
    plt.close()


def timestamp_vs_gasid(all_datetimes, all_idx, all_locs, output_dir=None, show=False):
    """
        ATTENTION: Most of this code is super ugly and will only work with the full dataset!
    Args:
        all_datetimes ([type]): [description]
        all_idx ([type]): [description]
        all_locs ([type]): [description]
        output_dir ([type], optional): [description]. Defaults to None.
        show (bool, optional): [description]. Defaults to False.
    """
    # Preprocess
    x, y, o = _get_nestedeventtimeslist_and_locs(all_datetimes, all_idx, all_locs)
    new_y = len(y) - np.array(y) + 1

    # Plot
    fig, ax = plt.subplots(figsize=(10, 4))
    ax.eventplot(
        positions=[[x[0][0]] for i in range(6)],
        lineoffsets=-5,
        colors=[[cm_inferno_r((_) / 6)] for _ in range(6)],
    )
    ax.legend([f"P{i+1}" for i in range(6)], loc="upper left")

    ax.eventplot(
        positions=x,
        lineoffsets=new_y,
        linewidths=0.5,
        colors=[[cm_inferno_r((__ - 1) / 6) for __ in _] for _ in o],
    )

    # Format
    xticks = ax.get_xticks()
    xticks = [str(mPlotDATEs.num2date(xtick))[:-14] for xtick in xticks]
    ax.set_xticklabels(xticks)
    yticks = ax.get_yticks()
    ax.set_yticks(new_y)
    yticks = [
        constants.GasNames[int(ytick)] for ytick in y
    ]  # ATTENTION: This is super ugly and will only work with the full dataset!
    ax.set_yticklabels(yticks)
    ax.set_ylim([min(y) - 0.6, max(y) + 0.6])
    ax.tick_params(axis="x", rotation=90)  # ax.set_xticklabels(a, rotation=90)

    # Save&Show
    if output_dir is not None:
        plt.savefig(
            output_dir.joinpath("all_gases_eventplot.png"),
            dpi=1800,
            bbox_inches="tight",
        )
        plt.savefig(output_dir.joinpath("1c_all_gases_eventplot.svg"), bbox_inches="tight")
    if show:
        plt.show()
    plt.close()


def time_vs_baseline(
    all_datetimes,
    all_baselines,
    all_gases,
    all_gas_ints,
    module,
    sensors,
    consider_resistance=True,
    subtracting_dc_offset=False,
    output_dir=None,
    show=False,
):
    """Plot time vs baseline values.

    Args:
        all_datetimes ([type]): [description]
        all_baselines ([type]): [description]
        all_gases ([type]): [description]
        all_gas_ints ([type]): [description]
        module ([type]): [description]
        sensors ([type], optional): [description]. Defaults to constants.all_sensors.
        consider_resistance (bool, optional): [description]. Defaults to True.
        subtracting_dc_offset (bool, optional): [description]. Defaults to False.
        output_dir ([type], optional): [description]. Defaults to None.
        show (bool, optional): [description]. Defaults to False.
    """
    # Preprocess
    date64 = np.datetime64("2000-01-01")
    timedelta = pd.to_timedelta("1 days 00:00:00")
    longgaps = []
    longgaps_start = []
    longgaps_end = []
    for j in range(len(all_datetimes) - 1):
        before_gap = all_datetimes[j]
        after_gap = all_datetimes[j + 1]
        if (after_gap - before_gap) > timedelta:
            longgaps.append([before_gap, after_gap])
            longgaps_start.append(before_gap)
            longgaps_end.append(after_gap)
    intervals_datetime = [[all_datetimes[0], longgaps[0][0]]]
    for j in range(len(longgaps) - 1):
        intervals_datetime.append([longgaps[j][1], longgaps[j + 1][0]])
    intervals_datetime.append([longgaps[-1][1], all_datetimes[-1]])
    intervals = (
        np.array(
            [[val.to_numpy() for val in interval] for interval in intervals_datetime]
        )
        - date64
    ) / np.timedelta64(1, "s")
    width_ratios = [interv[1] - interv[0] for interv in intervals]

    # Plot
    fig, axes = plt.subplots(
        nrows=2,
        ncols=len(width_ratios),
        sharey="row",
        sharex="col",
        gridspec_kw={"width_ratios": width_ratios, "height_ratios": [1, 12]},
        figsize=(15, 5),
    )

    for idx, sensor in enumerate(sensors):
        for datetime, df_baseline, gas, gas_int in zip(
            all_datetimes, all_baselines, all_gases, all_gas_ints
        ):
            for int_idx, interval in enumerate(intervals):
                datetime_seconds = (datetime.to_numpy() - date64) / np.timedelta64(
                    1, "s"
                )
                if (datetime_seconds >= interval[0]) and (
                    datetime_seconds <= interval[1]
                ):
                    axes[1, int_idx].scatter(
                        datetime,
                        df_baseline[f"B{module}S{sensors[idx]}"],
                        color=cm_viridis(idx / 7),
                        label=f"S{sensor}",
                        alpha=0.8,
                        s=10,
                        zorder=1,
                    )
                    axes[0, int_idx].scatter(
                        datetime,
                        65,
                        color=cm_rainbow(gas_int / 11),
                        s=10,
                        marker="s",
                        label=gas,
                    )

    # Format
    if consider_resistance:
        axes[1, 0].set_ylabel(r"$R_{sensor}$ [$k\Omega$]")
    else:
        axes[1, 0].set_ylabel(f"S{sensor} [mV]")
    for int_idx, interval in enumerate(intervals_datetime):
        axes[1, int_idx].set_xticks([interval[0], interval[1]])
        axes[1, int_idx].set_xticklabels([interval[0], interval[1]], rotation=90)
    for ax in axes[1, :]:
        xticks = ax.get_xticks()
        xticks = [str(mPlotDATEs.num2date(xtick))[:-9] for xtick in xticks]
        ax.set_xticklabels(xticks)
    for ax in axes[0, :]:
        ax.tick_params(left=False, bottom=True, labelleft=False, labelbottom=False)
    axes[0, 0].set_ylabel("Gas")
    unique_ax1 = _get_unique_labels(fig, axes[1, :])
    axes[1, -1].legend(*zip(*unique_ax1), bbox_to_anchor=(1, 0), loc="lower left")
    unique_ax0 = _get_unique_labels(fig, axes[0, :])
    axes[0, -1].legend(*zip(*unique_ax0), bbox_to_anchor=(1, 0), loc="lower left")
    plt.subplots_adjust(right=0.8, bottom=0.2)
    fig.align_labels()

    # Save&Show
    if output_dir is not None:
        plt.savefig(
            output_dir.joinpath(
                f"2a_baseline_{'dc' if not subtracting_dc_offset else 'gain'}_{'R' if consider_resistance else 'V'}_20s_.png"
            ),
            dpi=600,
            bbox_inches="tight",
        )
        plt.savefig(
            output_dir.joinpath(
                f"2a_baseline_{'dc' if not subtracting_dc_offset else 'gain'}_{'R' if consider_resistance else 'V'}_20s_.svg"
            ),
            bbox_inches="tight",
        )
    if show:
        plt.show()
    plt.close()


def pca_projections(
    pca,
    ts_start,
    X_train_all,
    y_train,
    subtracting_dc_offset=False,
    consider_resistance=True,
    output_dir=None,
    show=False,
):
    """Project data into PC-space and plot for different windows.

    Args:
        pca ([type]): [description]
        ts_start ([type]): [description]
        X_train_all ([type]): [description]
        y_train ([type]): [description]
        subtracting_dc_offset (bool, optional): [description]. Defaults to False.
        consider_resistance (bool, optional): [description]. Defaults to True.
        output_dir ([type], optional): [description]. Defaults to None.
        show (bool, optional): [description]. Defaults to False.
    """
    # Project in PC-Space and Plot
    fig_pca_common, axes_pca_common = plt.subplots(
        ncols=len(ts_start), nrows=1, figsize=(20, 4), sharex=True, sharey=True
    )
    axes_pca_common = axes_pca_common.reshape(-1)

    for t_start, [i, X_train] in zip(ts_start, enumerate(X_train_all)):
        x = pca.transform(X_train)  # Project data into this space
        # r_pc1, r_pc2 = pca.explained_variance_ratio_[:2]

        for l, val in zip(y_train, x):
            color = cm_rainbow(l / 11)
            axes_pca_common[i].scatter(
                val[0], val[1], color=color, label=l, marker=constants.markers[l], s=10
            )
            axes_pca_common[i].set_box_aspect(1)

        axes_pca_common[i].set_title(f"t={np.round(t_start/1000)}s")
        axes_pca_common[i].set_xlabel("PC1")
        axes_pca_common[0].set_ylabel("PC2")

    # Format
    lines = []
    labels = []
    for ax in fig_pca_common.axes:
        axLine, axLabel = ax.get_legend_handles_labels()
        lines.extend(axLine)
        labels.extend(axLabel)

    labels, lines = zip(*sorted(zip(labels, lines), key=lambda t: t[0]))
    unique = [
        (h, constants.GasNames[int(l)])
        for i, (h, l) in enumerate(zip(lines, labels))
        if l not in labels[:i]
    ]
    axes_pca_common[-1].legend(*zip(*unique), bbox_to_anchor=(1, 0), loc="lower left")
    plt.subplots_adjust(right=0.85)

    # Save&Show
    if output_dir is not None:
        plt.savefig(
            output_dir.joinpath(
                f"3ab_pca{'_gain' if subtracting_dc_offset else '_dc'}.png"
            ),
            dpi=600,
        )
        plt.savefig(
            output_dir.joinpath(
                f"3ab_pca{'_gain' if subtracting_dc_offset else '_dc'}.svg"
            )
        )
    if show:
        plt.show()
    plt.close()


def classification_accuracies(
    ts_start,
    all_accs_dc,
    all_accs_gain,
    skip_gas,
    classifier="SVM",
    consider_resistance=True,
    output_dir=None,
    show=False,
):
    """Plot classification accuracy vs time-of-window.

    Args:
        ts_start ([type]): [description]
        all_accs_dc ([type]): [description]
        all_accs_gain ([type]): [description]
        classifier (str, optional): [description]. Defaults to "SVM".
        consider_resistance (bool, optional): [description]. Defaults to True.
        output_dir ([type], optional): [description]. Defaults to None.
        show (bool, optional): [description]. Defaults to False.
    """
    # Preprocess
    accs_dc_mean = np.mean(all_accs_dc, axis=0)
    accs_gain_mean = np.mean(all_accs_gain, axis=0)

    # Plot
    fig, ax = plt.subplots()
    ax.axvline(20, linestyle="--", color="#d18080ff", zorder=0)
    ax.axhline(
        1.0 / (len(constants.GasIndices) - len(skip_gas)),
        linestyle="--",
        color="gray",
        zorder=1,
    )  # , label='random classification')
    for idx, t_start in enumerate(ts_start / 1000):
        n = all_accs_dc.shape[0]
        ax.scatter([t_start] * n, all_accs_dc[:, idx], c="k", s=3, zorder=2)
        ax.scatter([t_start] * n, all_accs_gain[:, idx], c="g", s=3, zorder=2)
    ax.plot(ts_start / 1000, accs_dc_mean, c="k", label="raw signal", zorder=3)
    ax.plot(ts_start / 1000, accs_gain_mean, c="g", label="gain", zorder=3)

    # Format
    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Validation Accuracy")
    ax.set_ylim([0, 1.05])
    ax.set_facecolor("#dce9f7")
    ax.legend(title="Linear SVM Classification")
    ax.grid()

    # Save&Show
    if output_dir is not None:
        plt.savefig(
            output_dir.joinpath(
                f"3cd_classification_{classifier}.png"
            ),
            dpi=600,
        )
        plt.savefig(
            output_dir.joinpath(
                f"3cd_classification_{classifier}.svg"
            )
        )
    if show:
        plt.show()
    plt.close()


def draw_windtunnel(
    values=None,
    output_dir=None,
    name=None,
    pos_x=[0.25, 0.5, 0.98, 1.18, 1.40, 1.45],
    pos_y=np.linspace(0, 1, 9) + 0.1,
    size_x=0.04,
    size_y=0.1,
    zlim=[None, None],
    show=False,
):
    """Draw heatmap of wind-tunnel

    Args:
        values ([type], optional): [description]. Defaults to None.
        output_dir ([type], optional): [description]. Defaults to None.
        name ([type], optional): [description]. Defaults to None.
        pos_x (list, optional): [description]. Defaults to [0.25, 0.5, 0.98, 1.18, 1.40, 1.45].
        pos_y ([type], optional): [description]. Defaults to np.linspace(0, 1, 9)+0.1.
        size_x (float, optional): [description]. Defaults to 0.04.
        size_y (float, optional): [description]. Defaults to 0.1.
        zlim (list, optional): [description]. Defaults to [None, None].
        show (bool, optional): [description]. Defaults to False.
    """
    # Preprocess
    if values is None:
        values = np.random.rand(len(pos_y), len(pos_x))
        values = np.arange(len(pos_x) * len(pos_y)).reshape(len(pos_y), len(pos_x))
    assert values.shape == (len(pos_y), len(pos_x))
    min_value = np.min(values)
    max_value = np.max(values)

    zmin, zmax = zlim
    if zmin is not None:
        min_value = zmin
        max_value = zmax

    # Plot
    fig, ax = plt.subplots()
    ax.plot([0], [0], alpha=1.0)
    for y, row in zip(pos_y, values):
        for x, val in zip(pos_x, row):
            if name != "indicator":
                ax.add_patch(
                    Rectangle(
                        (x - 0.5 * size_x, -y - 0.5 * size_y + 1.2),
                        size_x,
                        size_y,
                        color=cm_magma(_normalize(val, min_value, max_value)),
                    )
                )
            else:
                ax.add_patch(
                    Rectangle(
                        (x - 0.5 * size_x, -y - 0.5 * size_y + 1.2),
                        size_x,
                        size_y,
                        facecolor=cm_reds(_normalize(val, min_value, max_value)),
                        edgecolor="k",
                        linewidth=2,
                    )
                )

    # Format
    ax.set_xlim((0, 1.5))
    ax.set_ylim((0, 1.2))
    ax.set_xticks(pos_x)
    ax.set_xticklabels(["{:10.2f}".format(x) for x in pos_x], rotation=60)
    ax.set_xlabel("Longitudinal Position (m)")
    ax.set_ylabel("Lateral Position (m)")
    cbar = plt.colorbar(
        cm.ScalarMappable(
            norm=Normalize(vmin=min_value, vmax=max_value), cmap=cm_magma  # cm_inferno
        ),
        ax=ax,
        fraction=0.0222,
        pad=0.04,
    )
    cbar.ax.set_ylabel(f"{name}-term Drift")

    # Save&Show
    if output_dir is not None:
        plt.savefig(output_dir.joinpath(f"2bc_drift_windtunnel_{name}.png"), dpi=600)
        plt.savefig(output_dir.joinpath(f"2bc_drift_windtunnel_{name}.svg"))
    if show:
        plt.show()
    plt.close()


def indicate_windtunnel_location(
    location=4,
    modules=5,
    output_dir=None,
    name="indicator",
    pos_x=[0.25, 0.5, 0.98, 1.18, 1.40, 1.45],
    pos_y=np.linspace(0, 1, 9) + 0.1,
    size_x=0.04,
    size_y=0.1,
    show=False,
):
    """Plot simple wind-tunnel location indicator.

    Args:
        location (int, optional): [description]. Defaults to 4.
        modules (int, optional): [description]. Defaults to 5.
        output_dir ([type], optional): [description]. Defaults to None.
        name (str, optional): [description]. Defaults to "indicator".
        pos_x (list, optional): [description]. Defaults to [0.25, 0.5, 0.98, 1.18, 1.40, 1.45].
        pos_y ([type], optional): [description]. Defaults to np.linspace(0, 1, 9)+0.1.
        size_x (float, optional): [description]. Defaults to 0.04.
        size_y (float, optional): [description]. Defaults to 0.1.
        show (bool, optional): [description]. Defaults to False.
    """
    # Preprocess
    values = np.zeros((9, 6))
    for module in modules:
        values[module - 1, location - 1] = 1

    if values is None:
        values = np.random.rand(len(pos_y), len(pos_x))
        values = np.arange(len(pos_x) * len(pos_y)).reshape(len(pos_y), len(pos_x))
    assert values.shape == (len(pos_y), len(pos_x))
    min_value = np.min(values)
    max_value = np.max(values)

    # Plot
    fig, ax = plt.subplots(figsize=(3 * 1.5, 3 * 1.2))
    ax.plot([0], [0], alpha=1.0)
    for y, row in zip(pos_y, values):
        for x, val in zip(pos_x, row):
            ax.add_patch(
                Rectangle(
                    (x - 0.5 * size_x, -y - 0.5 * size_y + 1.2),
                    size_x,
                    size_y,
                    facecolor=cm_reds(_normalize(val, min_value, max_value)),
                    edgecolor="k",
                    linewidth=1,
                )
            )

    # Format
    ax.set_xlim((0, 1.5))
    ax.set_ylim((0, 1.2))
    ax.set_xticks([])
    ax.set_yticks([])

    # Save&Show
    if output_dir is not None:
        plt.savefig(output_dir.joinpath(f"windtunnel_{name}.png"), dpi=600)
        plt.savefig(output_dir.joinpath(f"windtunnel_{name}.svg"))
    if show:
        plt.show()


def draw_sensorboard(
    values=None, output_dir=None, name=None, zlim=[None, None], show=False
):
    """Draw heatmap of sensorboard.

    Args:
        values ([type], optional): [description]. Defaults to None.
        output_dir ([type], optional): [description]. Defaults to None.
        name ([type], optional): [description]. Defaults to None.
        zlim (list, optional): [description]. Defaults to [None, None].
        show (bool, optional): [description]. Defaults to False.
    """
    # Preprocess
    size = 0.4
    pos_x = [0, 1, 0, 1, 0, 1, 0, 1]
    pos_y = [3, 3, 2, 2, 1, 1, 0, 0]
    if values is None:
        values = np.random.rand(4, 2)
    values = np.reshape(values, (4, 2))
    min_value = np.min(values[values > 0])
    max_value = np.max(values[values > 0])

    zmin, zmax = zlim
    if zmin is not None:
        min_value = zmin
        max_value = zmax

    # Plot
    fig, ax = plt.subplots()
    ax.plot([0], [0], alpha=1.0)
    i = 0
    for row in values:
        for val in row:
            x = pos_x[i]
            y = pos_y[i]
            val_normalized = _normalize(val, min_value, max_value)
            ax.add_patch(
                Circle(
                    (x, y),
                    size,
                    color=cm_magma(val_normalized) if val != 0 else "w",
                    ec="k",
                )
            )
            ax.text(
                x,
                y,
                f"{i+1}",
                ha="center",
                va="center",
                color="k" if (val_normalized > 0.5 or val == 0) else "w",
            )
            i += 1

    # Format
    ax.set_aspect("equal", "box")
    ax.set_xticks([])
    ax.set_yticks([])
    cbar = plt.colorbar(
        cm.ScalarMappable(
            norm=Normalize(vmin=min_value, vmax=max_value), cmap=cm_magma
        ),
        ax=ax,
        fraction=0.0422,
        pad=0.04,
    )
    cbar.ax.set_ylabel(f"{name}-term Drift")

    # Save&Show
    if output_dir is not None:
        plt.savefig(output_dir.joinpath(f"drift_sensorboard_{name}.png"), dpi=600)
        plt.savefig(output_dir.joinpath(f"drift_sensorboard_{name}.svg"))
    if show:
        plt.show()
    plt.close()


def draw_boxplot_drift(values, output_dir=None, name=None, show=False):
    """Draw boxplot of drift for different sensor.

    Args:
        values ([type]): [description]
        output_dir ([type], optional): [description]. Defaults to None.
        name ([type], optional): [description]. Defaults to None.
        show (bool, optional): [description]. Defaults to False.
    """
    # Plot
    fig, ax = plt.subplots(figsize=(4, 3))
    ax.boxplot(
        values,
        widths=0.5,
        patch_artist=True,
        showmeans=False,
        boxprops=dict(facecolor="indigo", color="indigo"),
        medianprops=dict(color="yellow"),
        flierprops=dict(
            marker="o",
            markerfacecolor="w",
            markersize=2,
            linestyle="none",
            markeredgecolor="k",
        ),
        # showfliers = False,
        # showmeans=True
    )

    # Format
    ax.set_xticklabels(range(2, 9))
    ax.set_ylabel(f"{name}-term Drift")
    ax.set_xlabel("Sensor ID")
    plt.tight_layout()
    plt.grid(axis="x")

    # Save&Show
    if output_dir is not None:
        plt.savefig(
            output_dir.joinpath(f"2de_boxplot_{name}.png"), dpi=600, bbox_inches="tight"
        )
        plt.savefig(
            output_dir.joinpath(f"2de_boxplot_{name}.svg"), dpi=600, bbox_inches="tight"
        )
    if show:
        plt.show()
    plt.close()
